-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core module for TriviA-0-v2
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @author     Michael X. Lyons
--!             modified CAESAR API v1.0-3 VHDL for TriviA0v2
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.all;  -- needed for signal "checksum_number" of type "unsigned"
use work.AEAD_pkg.all;


entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 64;    --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    -- internal signals between Controller and Datapath
    signal ld_key          : std_logic;  -- load local key register from "key"
    signal ld_npub         : std_logic;  -- load "npub" from "bdi"

    signal wr_tag          : std_logic;  -- write the tag out as "bdo"
    signal ld_tag          : std_logic;  -- load a tag block from "bdi"
    signal tag_block1      : std_logic;  -- write/load the 1st block of the tag
    signal tag_ok          : std_logic;  -- the tag is as expected

    signal en_SC           : std_logic;  -- enable the stream cipher
    signal ld_SC           : std_logic;  -- initialize the stream cipher and load the key and IV
    signal en_count        : std_logic;  -- enable the cycle counter
    signal ld_count        : std_logic;  -- load the cycle counter with its start value
    signal count_done      : std_logic;  -- the cycle counter is at the required value
    signal en_block        : std_logic;  -- enable block processing
    signal ld_block        : std_logic;  -- this is the first block for checksum and tag processing
    signal block_limit     : std_logic;  -- we hit the block limit
    signal is_AD           : std_logic;  -- the input block is AD, not message/ciphertext
    signal is_decrypt      : std_logic;  -- the operation is decryption, not encryption
    signal is_checksum     : std_logic;  -- the input block is a checksum block, not AD or message/ciphertext
    signal checksum_number : unsigned(1 downto 0);  -- index (1, 2, 3) of the checksum portion to process as a block
    signal insert_tag      : std_logic;  -- insert the tag into the stream cipher state
begin
    cc_datapath: entity work.TriviA0v2_Datapath(dataflow)
        generic map (
            G_DBLK_SIZE  => G_DBLK_SIZE,
            G_KEY_SIZE   => G_KEY_SIZE,
            G_TAG_SIZE   => G_TAG_SIZE,
            G_LBS_BYTES  => G_LBS_BYTES
        )
        port map (
            clk              => clk,

            key              => key,

            bdi              => bdi,
            bdi_size         => bdi_size,
            bdi_pad_loc      => bdi_pad_loc,

            bdo              => bdo,
            bdo_size         => bdo_size,

            ld_key           => ld_key,
            ld_npub          => ld_npub,

            wr_tag           => wr_tag,
            ld_tag           => ld_tag,
            tag_block1       => tag_block1,
            tag_ok           => tag_ok,

            en_SC            => en_SC,
            ld_SC            => ld_SC,
            en_count         => en_count,
            ld_count         => ld_count,
            count_done       => count_done,
            en_block         => en_block,
            ld_block         => ld_block,
            block_limit      => block_limit,
            is_AD            => is_AD,
            is_decrypt       => is_decrypt,
            is_checksum      => is_checksum,
            checksum_number  => checksum_number,
            insert_tag       => insert_tag
        );

    cc_controller: entity work.TriviA0v2_Controller(behavioral)
        generic map (
            G_ASYNC_RSTN => G_ASYNC_RSTN,
            G_KEY_SIZE   => G_KEY_SIZE
        )
        port map (
            clk              => clk,
            rst              => rst,

            key_ready        => key_ready,
            key_valid        => key_valid,
            key_update       => key_update,

            decrypt          => decrypt,
            bdi_ready        => bdi_ready,
            bdi_valid        => bdi_valid,
            bdi_type         => bdi_type,
            bdi_eot          => bdi_eot,
            bdi_pad_loc      => bdi_pad_loc,

            bdo_valid        => bdo_valid,
            bdo_ready        => bdo_ready,

            msg_auth_done    => msg_auth_done,
            msg_auth_valid   => msg_auth_valid,

            ld_key           => ld_key,
            ld_npub          => ld_npub,

            wr_tag           => wr_tag,
            ld_tag           => ld_tag,
            tag_block1       => tag_block1,
            tag_ok           => tag_ok,

            en_SC            => en_SC,
            ld_SC            => ld_SC,
            en_count         => en_count,
            ld_count         => ld_count,
            count_done       => count_done,
            en_block         => en_block,
            ld_block         => ld_block,
            block_limit      => block_limit,
            is_AD            => is_AD,
            is_decrypt       => is_decrypt,
            is_checksum      => is_checksum,
            checksum_number  => checksum_number,
            insert_tag       => insert_tag
        );
 end structure;