-------------------------------------------------------------------------------
--! @file       TriviA0v2_Block.vhd
--! @brief      Block processing function from TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
--! Bit numbering follows the convention in the TriviA0v2 specification:
--!   bits are numbered from 1 up to the length.
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.GF2_32_package.all;
use work.GF2_64_package.all;


entity TriviA0v2_Block is
    generic (
        G_LBS_BYTES     : integer := 4
    );
    port (
         clk              : in  std_logic;                     -- clock
         en_block         : in  std_logic;                     -- enable block processing
         ld_block         : in  std_logic;                     -- this is the first block for checksum and tag processing
         block_limit      : out std_logic;                     -- we hit the block limit
         bdi              : in  std_logic_vector(1 to  64);    -- block to process
         bdi_size         : in  std_logic_vector(G_LBS_BYTES+1 -1 downto 0);  -- number of bytes in 'bdi'
         bdi_pad_loc      : in  std_logic_vector(1 to 8);      -- active bit shows position of padding (if any)
         is_AD            : in  std_logic;                     -- the input block is AD, not message/ciphertext
         is_decrypt       : in  std_logic;                     -- the operation is decryption, not encryption
         keyExt64         : in  std_logic_vector (1 to  64);   -- incoming value of keyExt64 (aka K^v "variable key")
         stExt64          : in  std_logic_vector (1 to  64);   -- incoming value of stExt64 (aka K^t "tag key")
         is_checksum      : in  std_logic;                     -- the input block is a checksum block, not AD or message/ciphertext
         checksum_number  : in  unsigned(1 downto 0);          -- index (1, 2, 3) of the checksum portion to process as a block
         bdo_out          : out std_logic_vector(1 to  64);    -- output block
         bdo_size         : out std_logic_vector(G_LBS_BYTES+1 -1 downto 0);  -- number of bytes in 'bdo'
         tag_out          : out std_logic_vector (1 to 128)    -- updated tag
         -- only for testing      checksum_out     : out std_logic_vector (1 to 192)   -- updated checksum
         );
end entity TriviA0v2_Block;

architecture behavioral of TriviA0v2_Block is
    -- block counter
    constant BLOCK_MAX_WIDTH : natural := 32;
    signal block_count : unsigned(BLOCK_MAX_WIDTH downto 0);  -- enough to hold 2^32
    constant BLOCK_COUNT_START : unsigned(BLOCK_MAX_WIDTH downto 0) := (BLOCK_MAX_WIDTH => '1',
                                                                        others          => '0');

    -- registers
    signal checksum_reg : std_logic_vector(1 to 192);
    alias cs1_reg  is checksum_reg(  1 to  64);
    alias cs2_reg  is checksum_reg( 65 to 128);
    alias cs3_reg  is checksum_reg(129 to 192);
    signal checksum_next : std_logic_vector(1 to 192);
    alias cs1_next is checksum_next(  1 to  64);
    alias cs2_next is checksum_next( 65 to 128);
    alias cs3_next is checksum_next(129 to 192);

    signal tag_reg  : std_logic_vector(1 to 128);
    alias tag1_reg  is tag_reg(  1 to  32);
    alias tag2_reg  is tag_reg( 33 to  64);
    alias tag3_reg  is tag_reg( 65 to  96);
    alias tag4_reg  is tag_reg( 97 to 128);
    signal tag_next : std_logic_vector(1 to 128);
    alias tag1_next is tag_next(  1 to  32);
    alias tag2_next is tag_next( 33 to  64);
    alias tag3_next is tag_next( 65 to  96);
    alias tag4_next is tag_next( 97 to 128);

    -- blocks in, out
    signal block_in  : std_logic_vector (1 to  64);
    signal block_xor : std_logic_vector (1 to  64);
    signal block_out : std_logic_vector (1 to  64);

    -- plaintext (bdi for AD or encryption, bdo for decryption)
    signal plaintext : std_logic_vector (1 to  64);
    
    -- double-bytes from 'block_in'
    alias x1 : std_logic_vector(1 to 16) is block_in( 1 to 16);
    alias x2 : std_logic_vector(1 to 16) is block_in(17 to 32);
    alias x3 : std_logic_vector(1 to 16) is block_in(33 to 48);
    alias x4 : std_logic_vector(1 to 16) is block_in(49 to 64);
    signal x1x3 : std_logic_vector(1 to 32);
    signal x2x4 : std_logic_vector(1 to 32);

    -- double-bytes from 'stExt64'
    alias k1 : std_logic_vector(1 to 16) is stExt64( 1 to 16);
    alias k2 : std_logic_vector(1 to 16) is stExt64(17 to 32);
    alias k3 : std_logic_vector(1 to 16) is stExt64(33 to 48);
    alias k4 : std_logic_vector(1 to 16) is stExt64(49 to 64);
    signal k1k3 : std_logic_vector(1 to 32);
    signal k2k4 : std_logic_vector(1 to 32);

    -- operands and result for finite field multiplication
    signal a_32 : std_logic_vector(1 to 32);  -- 32 bits from the input block and stream cipher state
    signal b_32 : std_logic_vector(1 to 32);  -- the other 32 bits from ...
    signal pdp  : std_logic_vector(1 to 32);  -- so-called "pseudo dot product"

begin
    --mult: entity work.Mult_in_GF2_32(interleaved)
    mult: entity work.Mult_in_GF2_32(mxl_mastrovito)
        port map (
            a  => a_32,
            b  => b_32,
            ab => pdp
        );

    -- registers
    process (clk)
    begin
        if rising_edge(clk) then
            if en_block = '1' then
                checksum_reg <= checksum_next;
                tag_reg      <= tag_next;

                if is_checksum /= '1' then
                    if ld_block = '1' then
                        block_count <= BLOCK_COUNT_START;
                    else
                        if block_count > 0 then
                            block_count <= block_count - 1;
                        end if;
                    end if;
                end if;
            end if;
        end if;
    end process;

    -- concurrent
    block_limit <= '1' when block_count = 2 else
                   '0';

    block_xor <= bdi xor keyExt64;

    -- if 'bdi' is padded then match the padding in 'block_out'
    with bdi_pad_loc select
        block_out <=  block_xor                           when "00000000",
                                           bdi( 1 to 64)  when "10000000",
                     (block_xor(1 to  8) & bdi( 9 to 64)) when "01000000",
                     (block_xor(1 to 16) & bdi(17 to 64)) when "00100000",
                     (block_xor(1 to 24) & bdi(25 to 64)) when "00010000",
                     (block_xor(1 to 32) & bdi(33 to 64)) when "00001000",
                     (block_xor(1 to 40) & bdi(41 to 64)) when "00000100",
                     (block_xor(1 to 48) & bdi(49 to 64)) when "00000010",
                     (block_xor(1 to 56) & bdi(57 to 64)) when "00000001",
                     (others => '0')                      when others;  -- should not happen

    bdo_out <= block_out;

    bdo_size <= bdi_size;

    plaintext <= bdi when (is_AD = '1') or (is_decrypt /= '1')
                     else block_out;

    --block_in <= cs1_reg when (is_checksum = '1') and (checksum_number = 1) else
    --            cs2_reg when (is_checksum = '1') and (checksum_number = 2) else
    --            cs3_reg when (is_checksum = '1') and (checksum_number = 3) else
    --            bdi     when (is_AD = '1') or (decrypt /= '1') else
    --            plaintext;
    -- this is 0.306s faster than the above
    block_in <= checksum_reg((((to_integer(checksum_number) - 1) * 64) + 1) to (((to_integer(checksum_number) - 1) * 64) + 64)) when is_checksum = '1'
                                                                                                                                else plaintext;

    -- set the operands for 32-bit finite field multiplication, called the "pseudo dot product" (sic) in the spec
    --a_32 <= (x1 & x3) xor (k1 & k3);
    --b_32 <= (x2 & x4) xor (k2 & k4);
    x1x3( 1 to 16) <= x1;
    x1x3(17 to 32) <= x3;
    x2x4( 1 to 16) <= x2;
    x2x4(17 to 32) <= x4;
    
    k1k3( 1 to 16) <= k1;
    k1k3(17 to 32) <= k3;
    k2k4( 1 to 16) <= k2;
    k2k4(17 to 32) <= k4;

    a_32 <= x1x3 xor k1k3;
    b_32 <= x2x4 xor k2k4;
    -- 'pdp' will be the product after a very short delay, make sure it's in the sensitivity list for tag processing below


    -- next-state logic
    -- ================
    -- checksum
    process (is_checksum,
             ld_block,
             block_in,
             checksum_reg)
    begin
        -- update the checksum with the input block
        
        -- add the block to the checksum if the block is not itself a checksum
        if is_checksum = '1' then
            checksum_next <= checksum_reg;
        else
            if ld_block = '1' then
                -- add the block to a zero checksum, i.e. the block value becomes the 3 parts of the checksum
                cs1_next <= block_in;
                cs2_next <= block_in;
                cs3_next <= block_in;
            else
                cs1_next <=              cs1_reg  xor block_in;
                cs2_next <= gf2_64_by_x( cs2_reg) xor block_in;
                cs3_next <= gf2_64_by_x2(cs3_reg) xor block_in;
            end if;
        end if;
    end process;

    -- tag
    process (ld_block,
             pdp,
             is_checksum,
             checksum_number,
             tag_reg,
             keyExt64)
        variable t1n, t2n, t3n, t4n : std_logic_vector(1 to 32);
    begin
        -- update the tag with the "pseudo dot product"
        
        -- avoid a latch if no path below updates any part of 'tag-next' (should not happen)
        tag_next <= tag_reg; 

        if ld_block = '1' then
            -- add the 'pdp' product to a zero tag, i.e. the 'pdp' value becomes the 4 parts of the tag
            tag1_next <= pdp;
            tag2_next <= pdp;
            tag3_next <= pdp;
            tag4_next <= pdp;
        else
            -- common elements of next values
            t1n :=              tag1_reg  xor pdp;
            t2n := gf2_32_by_x( tag2_reg) xor pdp;
            t3n := gf2_32_by_x2(tag3_reg) xor pdp;
            t4n := gf2_32_by_x4(tag4_reg) xor pdp;


            if (is_checksum = '1') and (checksum_number = 1) then
                tag1_next <= t1n xor keyExt64( 1 to 32);
                tag2_next <= t2n xor keyExt64(33 to 64);
            else
                tag1_next <= t1n;
                tag2_next <= t2n;
            end if;
            if (is_checksum = '1') and (checksum_number = 3) then
                tag3_next <= t3n xor keyExt64( 1 to 32);
                tag4_next <= t4n xor keyExt64(33 to 64);
            else
                tag3_next <= t3n;
                tag4_next <= t4n;
            end if;
                         
            --if is_checksum /= '1' then
            --    tag1_next <= t1n;
            --    tag2_next <= t2n;
            --    tag3_next <= t3n;
            --    tag4_next <= t4n;
            --else
            --    if    checksum_number = 1 then
            --        tag1_next <= t1n xor keyExt64( 1 to 32);
            --        tag2_next <= t2n xor keyExt64(33 to 64);
            --        tag3_next <= t3n;
            --        tag4_next <= t4n;
            --    elsif checksum_number = 2 then
            --        tag1_next <= t1n;
            --        tag2_next <= t2n;
            --        tag3_next <= t3n;
            --        tag4_next <= t4n;
            --    elsif checksum_number = 3 then
            --        tag1_next <= t1n;
            --        tag2_next <= t2n;
            --        tag3_next <= t3n xor keyExt64( 1 to 32);
            --        tag4_next <= t4n xor keyExt64(33 to 64);
            --    end if;
            --end if;
        end if;
    end process;

    -- outputs
    tag_out <= tag_reg;
    --        checksum_out <= checksum_reg;

end architecture behavioral;