-------------------------------------------------------------------------------
--! @file       TriviA0v2_Controller.vhd
--! @brief      Controller for CipherCore for TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.ALL;


entity TriviA0v2_Controller is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        G_DBLK_SIZE     : integer := 64;    --! Data
        G_KEY_SIZE      : integer := 128    --! Key
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3 -1 downto 0);
        bdi_eot         : in  std_logic;
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic;

        -- internal signals to/from Datapath
        ld_key          : out std_logic;  -- load local key register from "key"
        ld_npub         : out std_logic;  -- load "npub" from "bdi"

        wr_tag          : out std_logic;  -- write the tag out as "bdo"
        ld_tag          : out std_logic;  -- load a tag block from "bdi"
        tag_block1      : out std_logic;  -- write/load the 1st block of the tag
        tag_ok          : in  std_logic;  -- the tag is as expected

        en_SC           : out std_logic;  -- enable the stream cipher
        ld_SC           : out std_logic;  -- initialize the stream cipher and load the key and IV
        en_count        : out std_logic;  -- enable the cycle counter
        ld_count        : out std_logic;  -- load the cycle counter with its start value
        count_done      : in  std_logic;  -- the cycle counter is at the required value
        en_block        : out std_logic;  -- enable block processing
        ld_block        : out std_logic;  -- this is the first block for checksum and tag processing
        is_AD           : out std_logic;  -- the input block is AD, not message/ciphertext
        is_decrypt      : out std_logic;  -- the operation is decryption, not encryption
        block_limit     : in  std_logic;  -- we hit the block limit
        is_checksum     : out std_logic;  -- the input block is a checksum block, not AD or message/ciphertext
        checksum_number : out unsigned(1 downto 0);  -- index (1, 2, 3) of the checksum portion to process as a block
        insert_tag      : out std_logic   -- insert the tag into the stream cipher state
    );
end entity TriviA0v2_Controller;

architecture behavioral of TriviA0v2_controller is
    -- controller state
    type state_type is (
        S_WAIT_NPUB,   -- waiting for the npub for a new message/ciphertext
        S_WAIT_KEY,    -- waiting for a new key
        S_WAIT_LOAD,   -- wait one cycle to load npub or key
        S_START_1,     -- start counting cycles
        S_CYCLING_1A,  -- cycling the stream cipher state, allow one cycle for 'count_done' to clear from before
        S_CYCLING_1B,  -- cycling the stream state cipher before AD processing
        S_FIRST_AD,    -- waiting for the first AD block
        S_NEXT_AD,     -- waiting for the next  AD block
        S_CHKSUM_ADI1, -- processing checksum part 1 during AD
        S_CHKSUM_ADI2, -- processing checksum part 2 during AD
        S_CHKSUM_ADI3, -- processing checksum part 3 during AD
        S_CHKSUM_ADF1, -- processing checksum part 1 after AD
        S_CHKSUM_ADF2, -- processing checksum part 2 after AD
        S_CHKSUM_ADF3, -- processing checksum part 3 after AD
        S_INSERT,      -- inserting the current tag into stream cipher state
        S_START_2,     -- start counting cycles
        S_CYCLING_2A,  -- cycling the stream cipher state, allow one cycle for 'count_done' to clear from before
        S_CYCLING_2B,  -- cycling the stream state cipher before message/ciphertext processing
        S_FIRST_MC,    -- waiting for the first message/ciphertext block
        S_NEXT_MC,     -- waiting for the next  message/ciphertext block
        S_CHKSUM_MCI1, -- processing checksum part 1 during message/ciphertext
        S_CHKSUM_MCI2, -- processing checksum part 2 during message/ciphertext
        S_CHKSUM_MCI3, -- processing checksum part 3 during message/ciphertext
        S_CHKSUM_MCF1, -- processing checksum part 1 after message/ciphertext
        S_CHKSUM_MCF2, -- processing checksum part 2 after message/ciphertext
        S_CHKSUM_MCF3, -- processing checksum part 3 after message/ciphertext
        S_TAG_OUT1,    -- write the 1st block of the tag
        S_TAG_OUT2,    -- write the 2nd block of the tag
        S_TAG_VER1,    -- verify the 1st block of the tag
        S_TAG_VER2     -- verify the 2nd block of the tag
        );
    signal state_reg, state_next : state_type;  -- current, next state
    
    signal decrypt_reg, decrypt_next : std_logic;  -- internal copy of decrypt

begin
    -- reset and state change logic
    genResetSyncHi:  -- reset is synchronous, active high (typical of FPGAs)
    if not G_ASYNC_RSTN generate
        process(clk)
        begin
            if rising_edge(clk) then
                if rst = '1' then
                    state_reg <= S_WAIT_NPUB;
                else
                    state_reg <= state_next;
                end if;
            end if;
        end process;
    end generate;
    genResetAsyncLo:  -- reset is asynchronous, active low (typical of ASICs)
    if G_ASYNC_RSTN generate
        process(rst, clk)
        begin
            if rst = '0' then
                state_reg <= S_WAIT_NPUB;
            elsif rising_edge(clk) then
                state_reg <= state_next;
            end if;
        end process;
    end generate;


    -- registers
    process (clk)
    begin
        if rising_edge(clk) then
            decrypt_reg <= decrypt_next;
       end if;
    end process;


    -- next-state logic and data routing
    process (
            state_reg,
            -- inputs from PreProcessor
            key_update, key_valid,
            decrypt, decrypt_reg,
            bdi_valid, bdi_type, bdi_eot, bdi_pad_loc,
            -- inputs from PostProcessor
            bdo_ready,
            -- inputs from Datapath
            count_done,
            block_limit
            )
    begin
        -- default behavior:
        --   the state will stay the same, unless changed below
        state_next <= state_reg;

        --   registers hold their current value, unless set differently below
        decrypt_next <= decrypt_reg;

        --   output signals are cleared, unless set differently below
        key_ready     <= '0';
        bdi_ready     <= '0';
        bdo_valid     <= '0';
        msg_auth_done <= '0';

        ld_key <= '0';

        ld_npub <= '0';

        wr_tag     <= '0';
        ld_tag     <= '0';
        tag_block1 <= '0';

        en_SC       <= '0';
        ld_SC       <= '0';
        en_count    <= '0';
        ld_count    <= '0';
        en_block    <= '0';
        ld_block    <= '0';
        is_checksum <= '0';
        insert_tag  <= '0';
        --   not strictly necessary but will make the waveform display cleaner
        checksum_number <= to_unsigned(0, checksum_number'length);


        -- state check and possible change
        case state_reg is

            ----------------
            when S_WAIT_NPUB =>
                -- waiting for the npub for a new message/ciphertext

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for a new input

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_NPUB then  -- the input is "npub"
                        ld_npub <= '1';  -- tell the Datapath to save it

                        -- wait for the npub to be saved in a register
                        --   before loading the stream cipher state

                        decrypt_next <= decrypt;  -- store this because it will change while we are wrapping up

                        if key_update = '1' then  -- there is a new key
                            state_next <= S_WAIT_KEY;
                        else
                            state_next <= S_WAIT_LOAD;
                        end if;
                    end if;
                end if;

            ----------------
            when S_WAIT_KEY =>
                -- waiting for a new key

                key_ready <= '1';  -- tell the PreProcessor we are ready for the key

                if key_valid = '1' then  -- there is a new key available
                    ld_key <= '1';  -- tell the Datapath to save it
                    
                    -- wait for the key to be saved in a register
                    --   before loading the stream cipher state

                    state_next <= S_WAIT_LOAD;
                end if;

            ----------------
            when S_WAIT_LOAD =>
                -- wait one cycle to load npub or key

                en_SC <= '1';  -- enable the stream cipher
                ld_SC <= '1';  -- load key and npub into the stream cipher state

                state_next <= S_START_1;

            --------------
            when S_START_1 =>
                -- start counting cycles

                -- keep cycling
                en_SC <= '1';

                -- start counting
                en_count <= '1';  -- enable the counter
                ld_count <= '1';  -- load the counter with its IV

                state_next <= S_CYCLING_1A;

            ----------------
            when S_CYCLING_1A =>
                -- cycling the stream cipher state, allow one cycle for 'count_done' to clear from before

                -- keep cycling and counting
                en_SC    <= '1';
                en_count <= '1';

                state_next <= S_CYCLING_1B;

            ----------------
            when S_CYCLING_1B =>
                -- cycling the stream cipher state

                if count_done /= '1' then
                    -- keep cycling and counting
                    en_SC    <= '1';
                    en_count <= '1';
                else
                    state_next <= S_FIRST_AD;
                end if;

            --------------
            when S_FIRST_AD =>
                -- waiting for the first AD block

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for new data

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_ASS0 then  -- the input is AD
                        en_SC    <= '1';  -- enable the stream cipher
                        en_block <= '1';  -- enable block processing
                        ld_block <= '1';  -- this is the first block of AD (or the first after hitting the block limit)

                        if bdi_eot /= '1' then
                            -- there is at least one more AD block
                            state_next <= S_NEXT_AD;
                        else
                            -- this is the end of the AD blocks
                            state_next <= S_CHKSUM_ADF1;
                        end if;
                    end if;
                end if;

            --------------
            when S_NEXT_AD =>
                -- waiting for the next AD block

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for new data

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_ASS0 then  -- the input is AD
                        en_SC    <= '1';  -- enable the stream cipher
                        en_block <= '1';  -- enable block processing
                        ld_block <= '0';  -- this is NOT the first block of AD (nor the first after hitting the block limit)

                        if bdi_eot /= '1' then
                            -- there is at least one more AD block
                            if block_limit /= '1' then
                                -- no state change
                            else
                                -- process the current checksum then more AD blocks
                                state_next <= S_CHKSUM_ADI1;
                            end if;
                        else
                            -- this is the end of the AD blocks
                            state_next <= S_CHKSUM_ADF1;
                        end if;
                    end if;
                end if;

            -----------------
            when S_CHKSUM_ADI1 =>
                -- processing checksum part 1 during AD

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(1, checksum_number'length);

                state_next <= S_CHKSUM_ADI2;

            -----------------
            when S_CHKSUM_ADI2 =>
                -- processing checksum part 2 during AD

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(2, checksum_number'length);

                state_next <= S_CHKSUM_ADI3;

            -----------------
            when S_CHKSUM_ADI3 =>
                -- processing checksum part 3 during AD

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(3, checksum_number'length);

                state_next <= S_FIRST_AD;

            -----------------
            when S_CHKSUM_ADF1 =>
                -- processing checksum part 1 after AD

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(1, checksum_number'length);

                state_next <= S_CHKSUM_ADF2;

            -----------------
            when S_CHKSUM_ADF2 =>
                -- processing checksum part 2 after AD

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(2, checksum_number'length);

                state_next <= S_CHKSUM_ADF3;

            -----------------
            when S_CHKSUM_ADF3 =>
                -- processing checksum part 3 after AD

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(3, checksum_number'length);

                state_next <= S_INSERT;

            -------------
            when S_INSERT =>
                -- inserting the current tag into the stream cipher state

                -- insert the tag
                en_SC      <= '1';  -- enable the stream cipher
                insert_tag <= '1';  -- insert the tag into the stream cipher state

                state_next <= S_START_2;

            --------------
            when S_START_2 =>
                -- start counting cycles

                -- keep cycling
                en_SC <= '1';

                -- start counting
                en_count <= '1';  -- enable the counter
                ld_count <= '1';  -- load the counter with its IV

                state_next <= S_CYCLING_2A;

            ----------------
            when S_CYCLING_2A =>
                -- cycling the stream cipher state, allow one cycle for 'count_done' to clear from before

                -- keep cycling and counting
                en_SC    <= '1';
                en_count <= '1';

                state_next <= S_CYCLING_2B;

            ----------------
            when S_CYCLING_2B =>
                -- cycling the stream state cipher before message/ciphertext processing

                if count_done /= '1' then
                    -- keep cycling and counting
                    en_SC    <= '1';
                    en_count <= '1';
                else
                    state_next <= S_FIRST_MC;
                end if;

            --------------
            when S_FIRST_MC =>
                -- waiting for the first message/ciphertext block

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for new data

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_DAT0 then  -- the input is message/ciphertext
                        if bdi_pad_loc = "10000000" then
                            -- this block is just 10* padding, process it but don't write it out
                            -- this is the last block of message/ciphertext
                            en_SC    <= '1';  -- enable the stream cipher
                            en_block <= '1';  -- enable block processing
                            ld_block <= '1';  -- this is the first block of message/ciphertext (or the first after hitting the block limit)

                            state_next <= S_CHKSUM_MCF1;
                        else
                            if bdo_ready = '1' then
                                -- we can write to the output
                                bdo_valid <= '1'; -- tell the PostProcessor to take the output block

                                en_SC    <= '1';  -- enable the stream cipher
                                en_block <= '1';  -- enable block processing
                                ld_block <= '1';  -- this is the first block of message/ciphertext (or the first after hitting the block limit)

                                if bdi_eot /= '1' then
                                    -- there is at least one more message/ciphertext block
                                    state_next <= S_NEXT_MC;
                                else
                                    -- this is the end of the message/ciphertext blocks
                                    state_next <= S_CHKSUM_MCF1;
                                end if;
                            end if;
                        end if;
                    end if;
                end if;

            --------------
            when S_NEXT_MC =>
                -- waiting for the next message/ciphertext block

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for new data

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_DAT0 then  -- the input is message/ciphertext
                        if bdi_pad_loc = "10000000" then
                            -- this block is just 10* padding, process it but don't write it out
                            -- this is the last block of message/ciphertext
                            en_SC    <= '1';  -- enable the stream cipher
                            en_block <= '1';  -- enable block processing
                            ld_block <= '0';  -- this is NOT the first block of message/ciphertext (nor the first after hitting the block limit)

                            state_next <= S_CHKSUM_MCF1;
                        else
                            if bdo_ready = '1' then
                                -- we can write to the output
                                bdo_valid <= '1'; -- tell the PostProcessor to take the output block

                                en_SC    <= '1';  -- enable the stream cipher
                                en_block <= '1';  -- enable block processing
                                ld_block <= '0';  -- this is NOT the first block of message/ciphertext (nor the first after hitting the block limit)

                                if bdi_eot /= '1' then
                                    -- there is at least one more message/ciphertext block
                                    if block_limit /= '1' then
                                        -- no state change
                                    else
                                        state_next <= S_CHKSUM_MCI1;
                                    end if;
                                else
                                    -- this is the end of the message/ciphertext blocks
                                    state_next <= S_CHKSUM_MCF1;
                                end if;
                            end if;
                        end if;
                    end if;
                end if;

            -----------------
            when S_CHKSUM_MCI1 =>
                -- processing checksum part 1 during message/ciphertext

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number  <= to_unsigned(1, checksum_number'length);

                state_next <= S_CHKSUM_MCI2;

            -----------------
            when S_CHKSUM_MCI2 =>
                -- processing checksum part 2 after message/ciphertext

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(2, checksum_number'length);

                state_next <= S_CHKSUM_MCI3;

            -----------------
            when S_CHKSUM_MCI3 =>
                -- processing checksum part 3 after message/ciphertext

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(3, checksum_number'length);

                state_next <= S_FIRST_MC;

            -----------------
            when S_CHKSUM_MCF1 =>
                -- processing checksum part 1 after message/ciphertext

                -- $$$ do we wait for 'bdo_ready' to go high before we drop 'bdo_write_next'?

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number  <= to_unsigned(1, checksum_number'length);

                state_next <= S_CHKSUM_MCF2;

            -----------------
            when S_CHKSUM_MCF2 =>
                -- processing checksum part 2 after message/ciphertext

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(2, checksum_number'length);

                state_next <= S_CHKSUM_MCF3;

            -----------------
            when S_CHKSUM_MCF3 =>
                -- processing checksum part 3 after message/ciphertext

                en_SC       <= '1';  -- enable the stream cipher
                en_block    <= '1';  -- enable block processing
                is_checksum <= '1';  -- this block is a checksum, not AD or message/ciphertext
                checksum_number <= to_unsigned(3, checksum_number'length);

                if decrypt_reg /= '1' then
                    state_next <= S_TAG_OUT1;
                else
                    state_next <= S_TAG_VER1;
                end if;

            ---------------
            when S_TAG_OUT1 =>
                -- output the 1st block of the tag

                if bdo_ready = '1' then
                    -- tell Datapath to put the 1st block of the tag into "bdo"
                    wr_tag      <= '1';
                    tag_block1  <= '1';

                    bdo_valid <= '1'; -- tell the PostProcessor to take the output block

                    state_next <= S_TAG_OUT2;
                end if;

            ---------------
            when S_TAG_OUT2 =>
                -- output the 2nd block of the tag

                if bdo_ready = '1' then
                    -- tell Datapath to put the 2nd block of the tag into "bdo"
                    wr_tag      <= '1';
                    tag_block1  <= '0';

                    bdo_valid <= '1'; -- tell the PostProcessor to take the output block

                    -- that is it for this message, wait for the next
                    state_next <= S_WAIT_NPUB;
                end if;

            ---------------
            when S_TAG_VER1 =>
                -- verify the 1st block of the tag

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for new data

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_TAG then  -- the input is message/ciphertext
                        -- tell Datapath to put "bdi" into the first block of the expected tag
                        ld_tag     <= '1';
                        tag_block1 <= '1';

                        state_next <= S_TAG_VER2;
                    end if;
                end if;

            ---------------
            when S_TAG_VER2 =>
                -- verify the 2nd block of the tag

                bdi_ready <= '1';  -- tell the PreProcessor we are ready for new data

                if bdi_valid = '1' then  -- there is a new input available
                    if bdi_type = BDI_TYPE_TAG then  -- the input is message/ciphertext
                        -- tell Datapath to put "bdi" into the second block of the expected tag
                        ld_tag     <= '1';
                        tag_block1 <= '0';

                        -- tell the PreProcessor we have the verification result, and what it is
                        msg_auth_done  <= '1';

                        -- that is it for this message, wait for the next
                        state_next <= S_WAIT_NPUB;
                    end if;
                end if;

        end case;
    end process;


    -- outputs
    --   signals passed through from PreProcessor to Datapath
    is_decrypt <= decrypt_reg;

    is_AD  <= '1' when bdi_type = BDI_TYPE_ASS0
                  else '0';

    --   signals passed through from Datapath to PreProcessor
    msg_auth_valid <= tag_ok;

end architecture behavioral;