-------------------------------------------------------------------------------
--! @file       TriviA0v2_Datapath.vhd
--! @brief      Datapath for CipherCore for TriviA0v2
--! @project    CAESAR Candidate Evaluation
--! @author     Michael X. Lyons
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------


LIBRARY ieee;
USE ieee.std_logic_1164.all;
use ieee.numeric_std.all;


entity TriviA0v2_Datapath is
    generic (
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 64;    --! Data
        G_KEY_SIZE      : integer := 128;   --! Key
        G_TAG_SIZE      : integer := 128;   --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4
    );
    port (
        -- copies of internal signals exposed for testing  $$$ save this as ...extended.vhd when done
        --keyExt64_copy     : out std_logic_vector(1 to  64);  -- "variable key" out of stream cipher
        --stExt64_copy      : out std_logic_vector(1 to  64);  -- "state key" out of stream cipher
        --checksum_out_copy : out std_logic_vector(1 to 192);  -- updated checksum
        
        -- globals
        clk             : in  std_logic;

        -- from PreProcessor
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);

        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);

        -- to PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);

        -- internal signals from/to Controller
        ld_key          : in  std_logic;  -- load local key register from "key"
        ld_npub         : in  std_logic;  -- load "npub" from "bdi"

        wr_tag          : in  std_logic;  -- write the tag out as "bdo"
        ld_tag          : in  std_logic;  -- load a tag block from "bdi"
        tag_block1      : in  std_logic;  -- write/load the 1st block of the tag
        tag_ok          : out std_logic;  -- the tag is as expected

        en_SC           : in  std_logic;  -- enable the stream cipher
        ld_SC           : in  std_logic;  -- initialize the stream cipher and load the key and IV
        en_count        : in  std_logic;  -- enable the cycle counter
        ld_count        : in  std_logic;  -- load the cycle counter with its start value
        count_done      : out std_logic;  -- the cycle counter is at the required value
        en_block        : in  std_logic;  -- enable block processing
        ld_block        : in  std_logic;  -- this is the first block for checksum and tag processing
        block_limit     : out std_logic;  -- we hit the block limit
        is_AD           : in  std_logic;  -- the input block is AD, not message/ciphertext
        is_decrypt      : in  std_logic;  -- the operation is decryption, not encryption
        is_checksum     : in  std_logic;  -- the input block is a checksum block, not AD or message/ciphertext
        checksum_number : in  unsigned(1 downto 0);  -- index (1, 2, 3) of the checksum portion to process as a block
        insert_tag      : in  std_logic   -- insert the tag into the stream cipher state
    );
end entity TriviA0v2_Datapath;

architecture dataflow of TriviA0v2_Datapath is
    -- inputs
    signal key_reg, key_next : std_logic_vector(1 to G_KEY_SIZE);

    constant NPUB_SIZE : integer := G_DBLK_SIZE;   --! Npub size (bits)
    signal npub_reg, npub_next : std_logic_vector(1 to NPUB_SIZE);

    signal exp_tag_reg, exp_tag_next : std_logic_vector(1 to 128);  -- expected tag

    -- items for the stream cipher
    -- this is the 'param' value used in the TriviA0v2 reference code
    constant PARAM : std_logic_vector(1 to  64) := X"0000000000000000";

    signal iv             : std_logic_vector(1 to 128);  -- IV = param || npub
    signal keyExt64       : std_logic_vector(1 to  64);  -- "variable key" out of stream cipher
    signal stExt64        : std_logic_vector(1 to  64);  -- "state key" out of stream cipher

    -- items for counting stream cipher cycles
    constant NUM_LOOPS   : natural := 18;
    constant COUNT_LIMIT : natural := NUM_LOOPS - 1;  -- count starts at 0
    constant COUNT_WIDTH : natural := 5;  -- ceiling(log2(NUM_LOOPS))
    constant COUNT_IV    : unsigned((COUNT_WIDTH - 1) downto 0) := (others => '0');

    signal count_value    : unsigned((COUNT_WIDTH - 1) downto 0);

    -- items for block processing
    signal tag_out        : std_logic_vector(1 to 128);  -- updated tag

    -- output
    signal block_out : std_logic_vector(1 to  64);
    --###signal bdo_reg        : std_logic_vector(1 to  64);  -- 'bdo' output
    --###signal bdo_next       : std_logic_vector(1 to  64);  -- next value of 'bdo_reg'

    -- only for testing
    --signal checksum_out   : std_logic_vector(1 to 192);  -- updated checksum

begin
    -- combinatorial inputs
    --   IV = param || npub
    iv(  1 to  64) <= PARAM;
    iv( 65 to 128) <= npub_reg;

    exp_tag_reg( 65 to 128) <= bdi;

    -- registers
    process (clk)
    begin
        if rising_edge(clk) then
            -- update registers for inputs from PreProcessor
            key_reg <= key_next;

            npub_reg <= npub_next;

            exp_tag_reg(  1 to  64) <= exp_tag_next(  1 to  64);
       end if;
    end process;

    -- next-value logic for registers
    process (
            -- inputs from Controller
            ld_key, key, key_reg,
            ld_npub, bdi, npub_reg,
            ld_tag, tag_block1, exp_tag_reg
            )
    begin
        if ld_key = '1' then
            key_next <= key;
        else
            key_next <= key_reg;
        end if;

        if ld_npub = '1' then
            npub_next <= bdi;
        else
            npub_next <= npub_reg;
        end if;

        if (ld_tag = '1') and (tag_block1 = '1') then
            exp_tag_next(  1 to  64) <= bdi;
        else
            exp_tag_next(  1 to  64) <= exp_tag_reg(  1 to  64);
        end if;
    end process;

    -- stream cipher
    triviasc: entity work.TriviA0v2_SC(behavioral)
        port map (
            clk      => clk,
            en       => en_SC,
            load     => ld_SC,
            key      => key_reg,
            iv       => iv,
            insert   => insert_tag,
            tag      => tag_out,
            keyExt64 => keyExt64,
            stExt64  => stExt64
            );

    -- counter used for 18 cycles of stream state iteration
    loop_counter: entity work.counter_n_en_ld(behavioral)
        generic map (
            n => COUNT_WIDTH
        )
        port map (
            clk   => clk,
            en    => en_count,
            ld    => ld_count,
            d     => COUNT_IV,
            count => count_value
        );

    count_done <= '1' when count_value = COUNT_LIMIT
                      else '0';

    -- block processing
    triviablock: entity work.TriviA0v2_Block(behavioral)
        generic map (
            G_LBS_BYTES  => G_LBS_BYTES
        )
        port map (
            clk             => clk,
            en_block        => en_block,
            ld_block        => ld_block,
            block_limit     => block_limit,
            bdi             => bdi,
            bdi_size        => bdi_size,
            bdi_pad_loc     => bdi_pad_loc,
            is_AD           => is_AD,
            is_decrypt      => is_decrypt,
            keyExt64        => keyExt64,
            stExt64         => stExt64,
            is_checksum     => is_checksum,
            checksum_number => checksum_number,
            bdo_out         => block_out,
            bdo_size        => bdo_size,
            tag_out         => tag_out
            --             checksum_out    => checksum_out
            );

    -- outputs
    bdo <= block_out when wr_tag /= '1'
           else tag_out(  1 to  64) when tag_block1 = '1'
                else tag_out( 65 to 128);

    tag_ok <= '1' when tag_out = exp_tag_reg
              else '0';

    --keyExt64_copy     <= keyExt64;
    --stExt64_copy      <= stExt64;
    --checksum_out_copy <= checksum_out;

end architecture dataflow;